"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const jszip_1 = __importDefault(require("jszip"));
const StreamZip_1 = __importDefault(require("./StreamZip"));
const Errors_1 = require("../utils/Errors");
const events_1 = require("events");
const stream_1 = require("stream");
// @ts-ignore
const big_json_1 = __importDefault(require("big-json"));
const Settings_1 = __importDefault(require("../utils/Settings"));
function constructArchive(archive) {
    if (typeof archive === 'string' ||
        (typeof Blob !== 'undefined' && archive instanceof Blob)) {
        // Streaming method
        return new StreamArchive(archive);
    }
    else {
        // classic method
        return new Archive(archive);
    }
}
exports.constructArchive = constructArchive;
class StreamArchive {
    constructor(archive) {
        this._ready = Promise.resolve();
        this.current_dir = "";
        this.entries = {};
        this.events = new events_1.EventEmitter;
        if (archive instanceof StreamArchive) {
            this.s_zip = archive.s_zip;
            this.entries = Object.assign({}, archive.entries);
            this.current_dir = archive.current_dir;
            this._ready = archive._ready;
        }
        else {
            this.s_zip = new StreamZip_1.default(archive);
            this._ready = new Promise((resolve, reject) => {
                this.s_zip.on('ready', () => {
                    const s = this.s_zip;
                    // console.log('Entries read: ' + s.entriesCount);
                    this.entries = s.entries;
                    resolve();
                });
                this.s_zip.on('error', reject);
            });
        }
    }
    dir(dir_name) {
        const copy = new StreamArchive(this);
        if (dir_name.endsWith('/')) {
            dir_name = dir_name.slice(0, dir_name.length - 1);
        }
        const trimmed_entries = {};
        const name_regex = new RegExp('^' + dir_name + '/');
        const to_delete = [];
        for (const [name, entry] of Object.entries(copy.entries)) {
            if (!name.match(name_regex)) {
                to_delete.push(name);
            }
            trimmed_entries[this.ltrim(name, dir_name + "/")] = entry;
        }
        copy.entries = trimmed_entries;
        // Delete no valid entries
        for (const name of to_delete) {
            delete copy.entries[name];
        }
        return copy;
    }
    has(name) {
        return this.search(new RegExp(`^${name}$`)).length > 0;
    }
    search(query) {
        return Object.entries(this.files)
            .filter(f => !!f[0].match(query))
            .map(f => f[1]);
    }
    searchDir(query) {
        return Object.entries(this.dirs)
            .filter(f => f[0].match(query))
            .map(f => f[1]);
    }
    ready() {
        return this._ready;
    }
    ls(current_dir_only = true) {
        const l = this.entries;
        const files = {};
        let current_dir = this.current_dir;
        if (!current_dir) {
            current_dir = "";
        }
        for (const key in l) {
            if (key.startsWith(current_dir)) {
                const real_name = key.slice(current_dir.length);
                if (current_dir_only && real_name.match(/\/.+$/)) {
                    continue;
                }
                files[real_name] = l[key];
            }
        }
        return files;
    }
    get(name, type = "text", parse_auto = true) {
        const f = this.entries[name];
        if (!f) {
            this.events.emit('read error', { filename: name });
            throw new Errors_1.FileNotFoundError("File not found: " + name, name);
        }
        return this.read(f, type, parse_auto);
    }
    read(file, type = "text", parse_auto = true) {
        const fp = this.s_zip.entryData(file.name);
        function bufferToStream(binary) {
            const readableInstanceStream = new stream_1.Readable({
                read() {
                    this.push(binary);
                    this.push(null);
                }
            });
            return readableInstanceStream;
        }
        if (parse_auto) {
            return fp.then(data => {
                if (type === "text") {
                    let start_pos = 0;
                    let length = data.length;
                    for (let i = 0; i < length && i < 1024; i++) {
                        // Buffer code for "=" character
                        if (data[i] === 61) {
                            start_pos = i + 1;
                            break;
                        }
                    }
                    const buffer_part = data.slice(start_pos);
                    // > {LAZY_JSON_THRESHOLD} Mo
                    if (Settings_1.default.LAZY_JSON_PARSE && buffer_part.length > Settings_1.default.LAZY_JSON_THRESHOLD * 1024 * 1024) {
                        return new Promise((resolve, reject) => {
                            const stream_buffer = bufferToStream(buffer_part);
                            const parseStream = big_json_1.default.createParseStream();
                            parseStream.on('data', function (obj) {
                                // for an unknown reason, 
                                // arrays are not specified with the right prototype
                                if ('0' in obj) {
                                    // Setting prototype does not work,
                                    // So we're just getting an array from the object.
                                    resolve(Object.values(obj));
                                }
                                else {
                                    resolve(obj);
                                }
                            });
                            parseStream.on('error', function (err) {
                                reject(err);
                            });
                            stream_buffer.pipe(parseStream);
                        }).catch(e => {
                            if (e instanceof Error) {
                                throw new Errors_1.FileParseError(`Unexpected SyntaxError at JSON.parse when reading a file (${file.name}): ${e.message}`, file.name, "");
                            }
                            throw e;
                        });
                    }
                    const buffer_as_text = buffer_part.toString();
                    try {
                        return JSON.parse(buffer_as_text);
                    }
                    catch (e) {
                        if (e instanceof SyntaxError) {
                            throw new Errors_1.FileParseError(`Unexpected SyntaxError at JSON.parse when reading a file (${file.name}): ${e.message}`, file.name, buffer_as_text);
                        }
                        throw e;
                    }
                }
                const ab = new ArrayBuffer(data.length);
                const view = new Uint8Array(ab);
                for (let i = 0; i < data.length; ++i) {
                    view[i] = data[i];
                }
                if (type === "arraybuffer") {
                    return ab;
                }
                else {
                    return new Blob([ab]);
                }
            });
        }
        else {
            return fp;
        }
    }
    async fromFile(name) {
        let f;
        if (typeof name === 'string') {
            f = await this.get(name, "arraybuffer", false);
        }
        else {
            f = await this.read(name, "arraybuffer", false);
        }
        return new Archive(f);
    }
    get files() {
        const o = {};
        for (const [name, entry] of Object.entries(this.entries)) {
            if (!entry.isDirectory) {
                o[name] = entry;
            }
        }
        return o;
    }
    get dirs() {
        const o = {};
        for (const [name, entry] of Object.entries(this.entries)) {
            if (entry.isDirectory) {
                o[name] = entry;
            }
        }
        return o;
    }
    ltrim(name, left_str) {
        if (name.startsWith(left_str))
            return name.slice(left_str.length);
        return name;
    }
}
class Archive {
    constructor(file) {
        this._ready = Promise.resolve();
        this.events = new events_1.EventEmitter;
        if (file instanceof Archive) {
            this._ready = Promise.resolve();
            this.archive = file.archive;
            return;
        }
        if (file instanceof jszip_1.default) {
            this._ready = Promise.resolve();
            this.archive = file;
            return;
        }
        this._ready = jszip_1.default.loadAsync(file)
            .then(data => {
            this.archive = data;
        });
    }
    ready() {
        return this._ready;
    }
    dir(name) {
        return new Archive(this.archive.folder(name));
    }
    has(name) {
        return this.search(new RegExp(`^${name}$`)).length > 0;
    }
    get(name, type = "text", parse_auto = true) {
        if (!this.has(name)) {
            // @ts-ignore
            name = this.archive.root + name;
        }
        const f = this.archive.file(name);
        if (!f) {
            this.events.emit('read error', { filename: name });
            throw new Errors_1.FileNotFoundError("File not found: " + name, name);
        }
        return this.read(f, type, parse_auto);
    }
    search(query) {
        return this.archive.file(query);
    }
    searchDir(query) {
        return this.archive.folder(query);
    }
    read(file, type = "text", parse_auto = true) {
        const p = file.async(type);
        if (parse_auto) {
            return p.then(data => {
                if (typeof data === 'string') {
                    return JSON.parse(data.substr(data.indexOf('=') + 1).trimLeft());
                }
                else {
                    return data;
                }
            });
        }
        else {
            return p;
        }
    }
    ls(current_dir_only = true) {
        const l = this.archive.files;
        const files = {};
        // @ts-ignore
        let current_dir = this.archive.root;
        if (!current_dir) {
            current_dir = "";
        }
        for (const key in l) {
            if (key.startsWith(current_dir)) {
                const real_name = key.slice(current_dir.length);
                if (current_dir_only && real_name.match(/\/.+$/)) {
                    continue;
                }
                files[real_name] = l[key];
            }
        }
        return files;
    }
    /**
     * Create a new instance of Archive from a file contained in this archive.
     *
     * @param name File name or file object
     */
    async fromFile(name) {
        let f;
        if (typeof name === 'string') {
            f = await this.get(name, "arraybuffer", false);
        }
        else {
            f = await this.read(name, "arraybuffer", false);
        }
        return new Archive(f);
    }
    get raw() {
        return this.archive;
    }
}
exports.Archive = Archive;
exports.default = Archive;
